#ifndef lint
static char Rcs_Id[] =
    "$Header$";
#endif

/*
 * $Log$
 * Revision 2.23  2000/07/03 16:37:58  johnh
 * work-around for haning bugs
 *
 * Revision 2.22  2000/04/14 01:17:54  johnh
 * patch for y2k problem in at jobs
 *
 * Revision 2.21  2000/04/13 03:19:35  johnh
 * fixes for several bugs: loc-specific at jobs running multiple times and at-job reading, and geoffs mis-applied Y2K patch
 *
 * Revision 2.20  2000/03/23 01:32:15  geoff
 * Do the previous fix right (before, it broke queueing of jobs).
 *
 * Revision 2.19  2000/03/23 01:00:25  geoff
 * Fix a bug that caused at -r to fail for jobs after the 1900's.
 *
 * Revision 2.18  1999/07/27 01:21:59  johnh
 * location support for at jobs
 *
 * Revision 2.17  1999/07/14 05:22:30  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 2.16  1998/07/30 20:42:21  geoff
 * Many changes, accumulated over far too much time: Y2K fixes, support
 * -v switch, be careful about overrunning arrays, and various security
 * enhancements.
 *
 * Revision 2.15  1995/11/27  01:59:17  geoff
 * Require an ID in removal only to disambiguate.  List jobs in sorted
 * order.
 *
 * Revision 2.14  1995/05/11  05:04:57  geoff
 * Support a variable signal type.  Don't expect NULL as a return from
 * strcmp!
 *
 * Revision 2.13  88/10/31  23:51:17  geoff
 * Fix a couple of spelling typos in comments.
 * 
 * Revision 2.12  87/01/10  19:12:52  geoff
 * Make changes to enhance portability.  Don't complain if the cron daemon
 * isn't running.
 * 
 * Revision 2.11  86/11/30  00:23:41  geoff
 * Modify listjobs() to list the whole job, not just the first command.  Also
 * fix a longstanding bug that omitted the last newline of every at job.
 * 
 * Revision 2.10  86/11/29  19:49:25  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.9  86/11/16  00:56:43  geoff
 * Add support for using O_EXCL in temp file opens if the system has fcntl.h.
 * Run the user's shell with exec to save a process.
 * 
 * Revision 2.8  86/04/17  21:54:07  geoff
 * Add permit/deny support.
 * 
 * Revision 2.7  86/04/13  13:30:55  geoff
 * Add support for -r and -l.  Also make the zulu stuff work.
 * 
 * Revision 2.6  86/04/02  00:00:31  geoff
 * Make the here-document string be a #define in the header file.
 * 
 * Revision 2.5  86/03/30  14:32:20  geoff
 * Add (untested) code to support Zulu time (GMT).  This includes (untested)
 * code in incrby to allow negative increments.
 * 
 * Revision 2.4  86/03/28  14:40:07  geoff
 * Fix some lint complaints, plus a couple of small bugs (notably in the
 * handling of single quotes in environment variables.
 * 
 * Revision 2.3  86/03/25  19:54:40  geoff
 * Add code to actually set up the at job and pass it on to the daemon.
 * This also adds checkwhen to do error-checking.
 * 
 * Revision 2.2  86/01/19  13:35:31  geoff
 * Add 'fixwhen' and 'incrby' to apply the defaulting and incrementing rules.
 * Also add a day-of-week calculation to the debugging printout.
 * 
 * Revision 2.1  86/01/04  20:04:50  geoff
 * Initial revision
 * 
 */

#define GEN_ID				/* Generate source-control ID's */

#include "at.h"
#include "cron.h"
#include "ndir.h"
#include <errno.h>
#ifdef HASFCNTL
#include <fcntl.h>
#endif
#include <pwd.h>
#include <stdio.h>
#include <time.h>
#include <sys/signal.h>
#include <assert.h>

int		main ();	/* Execute commands at a later time */
static int	queue_or_rm ();	/* Queue up or remove an at job */
static void	usage ();	/* Issue a usage message */
static int	checkwhen ();	/* Check validity of 'when' */
static void	fixwhen ();	/* Fix unspecified fields in 'when' */
static void	incrby ();	/* Increment a time by a specified value */
static int	addjob ();	/* Add an at job to the queue */
static void	fprefix ();	/* Put out the at-job file prefix */
static void	fsuffix ();	/* Put out the at-job file suffix */
static SIG_TYPE	catchsigs ();	/* Catch unfriendly signals and clean up */
static int	listjobs ();	/* List all at jobs belonging to user */
static int	comparejobs();	/* Compare two jobs in sorting the job list */
static int	remjob ();	/* Remove an at job from the queue */

#ifndef YYDEBUG
#define YYDEBUG 0
#endif

/*
 * Structure used to store job names so that they can be sorted during
 * listjobs().
 */
struct jobname
    {
    char *	name;		/* Name of the job */
    int		uid;		/* UID of user who owns this job */
    };

char *		atstr = NULL;	/* Command string to be parsed */
int		incrcode;	/* Code for incrementation unit */
int		incrval;	/* Number of units to increment by */
struct tm *	now;		/* What time it is now */
char *		progname;	/* Name we are running under */
static int	verbose = 0;	/* NZ for verbose job listings */
struct tm	when;		/* When to run the command */
/* #ifdef LOC_CRON */
/* #ifdef LOC_ATLOC */
char *		where;		/* where to run the command */
/* #endif /* LOC_ATLOC */
/* #endif /* LOC_CRON */
static char	whenstr[] =	/* ASCII encoding of when to run */
			  WHEN_PROTOTYPE;
int		zulu = 0;	/* NZ if time is 'zulu' (GMT) */

static short	mlens[] =	/* Table of month lengths */
    {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

/*
 * Leap-year calculation.  This will work for all dates in the future until
 * the year 2100, by which time I hope this software is thoroughly obsolete.
 * Also note that it works whether applied to a 4-digit year or a Unix
 * "year - 1900".
 */
#define leap(year)	(((year) & 3) == 0)

/*
 * Leap-year adjustment calculation.  Generates 1 if month is February and
 * the year is a leap year.  Note months are indexed from zero.
 */
#define leapadj(month, year)	((month) == 1  &&  leap (year))

extern int	daylight;	/* NZ if system uses daylight-savings time */
extern time_t	dstamount ();	/* Calculate daylight-savings offset */
extern char **	environ;	/* Environment 'at' was invoked with */
extern char *	errmsg ();	/* Return error message string */
extern char *	getenv ();	/* Return value of environment variable */
extern char *	getname ();	/* Get user's name */
extern struct passwd *
		getpwuid ();	/* Get a password entry by uid */
extern char *	malloc ();	/* Allocate memory */
extern char *	optarg;		/* Command argument option */
extern int	optind;		/* Option index */
#if YYDEBUG
extern int	yydebug;	/* Debugging control */
#endif
extern int	sendcommand ();	/* Send a command to the cron daemon */
extern void	setlibdir ();	/* Change directories to LIBDIR */
extern char *	strcpy ();	/* Copy strings */
extern char *	tempfile ();	/* Create a temporary file */
extern time_t	time ();	/* Return system time */
extern long	timezone;	/* System timezone, in seconds west of GMT */
extern time_t	tm_to_time ();	/* Convert struct tm into time-of-day */
extern struct passwd * userpwent; /* User's password entry */
				/* Set by getname */

int main (argc, argv)			/* Execute commands later */
    reg2 int		argc;		/* Argument count */
    reg1 char *		argv[];		/* Argument vector */
    {
    int			list = 0;	/* NZ to list user's at jobs */
    int			optch;		/* Option char from getopt */
    int			remove = 0;	/* NZ to remove an at job */

    progname = argv[0];
#if YYDEBUG
    while ((optch = getopt (argc, argv, "dlrv")) != EOF)
#else
    while ((optch = getopt (argc, argv, "lrv")) != EOF)
#endif
	{
	switch (optch)
	    {
#if YYDEBUG
	    case 'd':
		yydebug = 1;		/* Turn on debugging */
		break;
#endif
	    case 'l':
		list = 1;
		break;
	    case 'r':
		remove = 1;
		break;
	    case 'v':
		verbose = 1;
		break;
	    default:
		usage ();
		return 1;
	    }
	}
    if ((list  &&  optind != argc)
      ||  (!list  &&  optind >= argc)
      ||  (!list  &&  verbose)
      ||  (list && remove))
	{
	usage ();
	return 1;
	}
    if (list)
	return listjobs ();
    else
	return queue_or_rm (argc, argv, remove);
    }

static int queue_or_rm (argc, argv, remove) /* Queue or remove an at job */
    reg3 int		argc;		/* Argument count */
    reg1 char *		argv[];		/* Argument vector */
    int			remove;		/* NZ to remove the job */
    {
    reg2 int		argind;		/* Index of time args */
    reg4 int		len;		/* Length of all time args */
    time_t		timenow;	/* The time right now */
    char *		user;		/* Name of user submitting job */

    (void) time (&timenow);		/* Set the default time */
    now = localtime (&timenow);		/* ... */
    when.tm_min = -1;			/* Make a recognizably illegal value */
    when.tm_hour = -1;			/* ..in all of when's useful fields */
    when.tm_mday = -1;
    when.tm_mon = -1;
    when.tm_year = -1;
    when.tm_wday = -1;
    incrcode = -1;
    incrval = 0;
/* #ifdef LOC_CRON */
/* #ifdef LOC_ATLOC */
    where = NULL;
/* #endif \* LOC_ATLOC */
/* #endif \* LOC_CRON */

    /*
     * If this is a remove request, the last argument is the ID of the
     * job to remove, so we don't include it in the string.  The following
     * statement ensures that remove is exactly 1 if it is nonzero.
     */
    if (remove)
	remove = 1;
    for (argind = optind, len = 0;  argind < argc - remove;  argind++)
	len += strlen (argv[argind]) + 1;
    if (!remove  &&  len == 0)
	{
	(void) fprintf (stderr, "at:  No time given\n");
	return 1;
	}
    if (len != 0)
	{
	atstr = malloc ((unsigned) len); /* Make space for the string */
	if (atstr == NULL)
	    {
	    (void) fprintf (stderr,
	      "Unable to allocate %d byte%s for arguments\n",
	      len, len == 1 ? "" : "s");
	    return 1;
	    }
	}
    for (argind = optind, len = 0;  argind < argc - remove;  argind++)
	{
	(void) strcpy (&atstr[len], argv[argind]);
	len += strlen (&atstr[len]);
	atstr[len++] = ' ';	/* Put blanks between arguments */
	}
    if (len == 0)
	{
	when.tm_sec = -1;
	when.tm_min = -1;
	when.tm_hour = -1;
	when.tm_mday = -1;
	when.tm_mon = -1;
	when.tm_year = -1;
	when.tm_wday = -1;
	when.tm_yday = -1;
	when.tm_isdst = -1;
	}
    else
	{
	atstr[--len] = '\0';	/* Change last blank to a null byte */
	if (yyparse ())		/* Parse the arguments */
	    return 1;		/* yyparse puts out syntax error message */
	}
#ifdef LOC_CRON
    if (!where && 
	/* when.tm_sec == -1 && */
	when.tm_min == -1 &&
	when.tm_hour == -1 &&
	when.tm_mday == -1 &&
	when.tm_mon == -1 &&
	when.tm_year == -1 &&
	when.tm_wday == -1 &&
	/* when.tm_yday == -1 && */
	/* when.tm_isdst == -1 && */
	1) {
	    (void) fprintf (stderr, "must specify either location or time.\n");
	    return 1;
    };
#else
    if (where) {
	    (void) fprintf (stderr, "location specified, but at/cron not compiled with location support.\n");
	    return 1;
    };
    if (/* when.tm_sec == -1 && */
	when.tm_min == -1 &&
	when.tm_hour == -1 &&
	when.tm_mday == -1 &&
	when.tm_mon == -1 &&
	when.tm_year == -1 &&
	when.tm_wday == -1 &&
	/* when.tm_yday == -1 && */
	/* when.tm_isdst == -1 && */
	1) {
	    (void) fprintf (stderr, "must specify either some time.\n");
	    return 1;
    };
#endif /* LOC_CRON */
    if (zulu)
	now = gmtime (&timenow); /* Switch to GMT defaulting rules */

    if (remove)
	return remjob (argv[argc - 1], len == 0);

    if (!allowed (user, ATPRIV))
	{
	(void) fprintf (stderr,
	  "Sorry, you are not allowed to submit at jobs\n");
	return 2;
	}

    if (checkwhen (1))
	return 1;		/* Illegal date */
    fixwhen ();			/* Fill in unspecified fields in 'when' */
#if YYDEBUG
    if (yydebug)
	{
	static short adj[12] = {1, 4, 4, 0, 2, 5, 0, 3, 6, 1, 4, 6};
	when.tm_wday =
	  when.tm_year / -50		/* CC */
	  + ((when.tm_year % 100) * 5) / 4	/* + 1.25 * YY */
	  + adj[when.tm_mon]		/* + months[mon] */
	  + when.tm_mday			/* + dd */
	  - (leap (when.tm_year)  &&  when.tm_mon < 2); /* - fudge */
	when.tm_wday = when.tm_wday % 7;
	(void) printf ("Scheduled for %s", asctime (&when));
	}
#endif

    if (tm_to_time (&when) < timenow - CATCHUP)
	{
	(void) fprintf (stderr, "at: too late\n");
	return 1;
	}

    user = getname (NULL);
    if (user == NULL)
	return 1;

    return addjob ();
    }

static void usage ()		/* Print a usage message */
    {
#if YYDEBUG
    (void) fprintf (stderr, "Usage:\tat [-d] time [date] [incr]\n");
    (void) fprintf (stderr, "\tat -l [-v]\n");
    (void) fprintf (stderr, "\tat -r time [date] [incr] ID\n");
#else
    (void) fprintf (stderr, "Usage:\tat time [date] [incr]\n");
    (void) fprintf (stderr, "\tat -l [-v]\n");
    (void) fprintf (stderr, "\tat -r time [date] [incr] ID\n");
#endif
    }

static int checkwhen (needhm)	/* Check 'when' for legality */
    int		needhm;		/* NZ to insist on having hour/minute */
    {				/* ..returns nonzero if problems */
#ifdef LOC_CRON
#ifdef LOC_ATLOC
    if (where)
        return 0;
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */
    if ((needhm  ||  when.tm_min != -1)
      &&  (when.tm_min < 0  ||  when.tm_min >= 60))
	{
	(void) fprintf (stderr, "Illegal minute\n");
	return 1;
	}
    else if ((needhm  ||  when.tm_hour != -1)
      &&   (when.tm_hour < 0  ||  when.tm_hour > 24
	||  (!zulu  &&  when.tm_hour == 24)))
	{
	(void) fprintf (stderr, "Illegal hour\n");
	return 1;
	}
    if (when.tm_hour == 24)
	when.tm_hour = 0;	/* In zulu, 24:xx is the same as 0:xx */
    /*
     * If a weekday was specified, its validity is assured by the parser.
     * Similarly, the validity of the month is guaranteed.
     */
    if (when.tm_mon != -1)	/* Month/day specified? */
	{
	/*
	 * Checking the day of the month is complicated by varying month
	 * lengths and leap years.  Further, if the year is defaulted, the
	 * year could be either this year or next depending on whether the
	 * date is less than today's.  This is all taken care of by
	 * special-casing February 29th.
	 */
	if (when.tm_mon == 1)	/* February? */
	    {
	    if (when.tm_mday == 29)
		{
		/*
		 * The user wants February 29th.  If the year is defaulted,
		 * he could mean this year or next.  If this year is leap
		 * year, then Feb 29th is legal only if it is later than
		 * now (taking time into account down to the hour level).
		 * If next year is leap year, Feb 29th is legal only if it
		 * earlier, by the same criterion.
		 */
		if (when.tm_year != -1)
		    {
		    if (!leap (when.tm_year))
			when.tm_mday = 99;	/* Produce an illegal value */
		    }
		else if (now->tm_mon < 1
		  ||  (now->tm_mon == 1  &&  now->tm_mday < 29)
		  ||  (now->tm_mon == 1  &&  now->tm_mday == 29
		    &&  now->tm_hour < when.tm_hour)
		  ||  (now->tm_mon == 1  &&  now->tm_mday == 29
		    &&  now->tm_hour == when.tm_hour
		    &&  now->tm_min < when.tm_min))
		    {
		    if (!leap (now->tm_year))
			when.tm_mday = 99;	/* Produce an illegal value */
		    }
		else				/* Date is in next year */
		    {
		    if (!leap (now->tm_year + 1))
			when.tm_mday = 99;	/* Produce an illegal value */
		    }
		}
	    if (when.tm_mday <= 0  ||  when.tm_mday > 29)
		{
		(void) fprintf (stderr, "Illegal day\n");
		return 1;
		}
	    }
	else if (when.tm_mday <= 0  ||  when.tm_mday > mlens[when.tm_mon])
	    {
	    (void) fprintf (stderr, "Illegal day\n");
	    return 1;
	    }
	}
    /*
     * All years are legal from our point of view, though some will be
     * considered too late.
     */
    return 0;
    }

static void fixwhen ()		/* Fix unspecified fields in 'when' */
    {
    /*
     * The time specification from the user is of the form
     * "time date increment".  The increment unit is stored in the external
     * 'incrcode' and the number of units in 'incrval';  the rest of the
     * specification is stored in 'when'.
     *
     * The time can be an hour/minute specification, or it can be one
     * of the special keywords 'now', 'midnight', or 'noon'.
     *
     * The date can be omitted entirely, or it can one of the following:
     *	    a weekday name
     *	    a month name followed by a day number
     *	    a month name, day number, and year
     *
     * The defaulting rules are as follows:
     *
     *	    There is no default time.
     *	    Unless this job is only location-based and
     *		there is no time specification.
     *	    If the date is entirely omitted, the default day is today unless
     *		the time is less than (not <=) the time right now.  If the
     *		time given is less than the current time, the default day is
     *		tomorrow.
     *	    If the date is a weekday, it is the next occurrence of that
     *		weekday.  If the weekday is the same as today, the date
     *		referred to is next week.
     *	    If the date is a month/day combination without a year, the same
     *		algorithm is applied:  it refers to the next occurrence of
     *		that date.  If it is the same as today's date, it refers to
     *		next year.
     *
     * When deciding whether to use the current day, month, or year, ties
     * are broken by the time taken to the hour/minute level.
     *
     * Once the defaults have been applied, the increment is applied.
     */
#ifdef LOC_CRON
#ifdef LOC_ATLOC
    if (where && when.tm_min == -1 && when.tm_hour == -1 &&
	when.tm_mday == -1 && when.tm_mon == -1 && when.tm_year == -1 &&
	when.tm_wday == -1) {
	    /*
	     * only where => run job ASAP 
	     */
	    when.tm_hour = now->tm_hour;
	    when.tm_min = now->tm_min;
    };
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */
    if (when.tm_wday == -1  &&  when.tm_mon == -1) /* No date at all? */
	{
	when.tm_mday = now->tm_mday;
	when.tm_mon = now->tm_mon;
	when.tm_year = now->tm_year + 1900;
	if (when.tm_hour < now->tm_hour
	  ||  (when.tm_hour == now->tm_hour  &&  when.tm_min < now->tm_min))
	    incrby (&when, I_DAY, 1);	/* Default is tomorrow */
	}
    else if (when.tm_wday != -1)	/* Weekday specified? */
	{
	when.tm_mday = now->tm_mday;
	when.tm_mon = now->tm_mon;
	when.tm_year = now->tm_year;
	if (when.tm_wday < now->tm_wday
	  ||  (when.tm_wday == now->tm_wday  &&  when.tm_hour < now->tm_hour)
	  ||  (when.tm_wday == now->tm_wday  &&  when.tm_hour == now->tm_hour
	    &&  when.tm_min < now->tm_min))
	    incrby (&when, I_DAY, when.tm_wday + 7 - now->tm_wday);
	else
	    incrby (&when, I_DAY, when.tm_wday - now->tm_wday);
	}
    else if (when.tm_year == -1)	/* Month/day specified, but no year? */
	{
	when.tm_year = now->tm_year;
	if (when.tm_mon < now->tm_mon
	  ||  (when.tm_mon == now->tm_mon  &&  when.tm_mday < now->tm_mday)
	  ||  (when.tm_mon == now->tm_mon  &&  when.tm_mday == now->tm_mday
	    &&  when.tm_hour < now->tm_hour)
	  ||  (when.tm_mon == now->tm_mon  &&  when.tm_mday == now->tm_mday
	    &&  when.tm_hour == now->tm_hour  &&  when.tm_min < now->tm_min))
	    incrby (&when, I_YEAR, 1);	/* Default is next year */
	}
    /*
     * Fix 2-digit dates to be 4-digit
     */
    if (when.tm_year < Y2K_THRESHOLD)
	when.tm_year += Y2K_2000;
    if (incrval)
	incrby (&when, incrcode, incrval);
    if (zulu)
	{
	/*
	 * The unhelpful user has chosen to specify 'when' in GMT.  Now
	 * we have to convert it to local time, complete with DST.
	 */
	tzset ();
	incrby (&when, I_MINUTE, (int) (-timezone / MINUTE));
	if (daylight)
	    incrby (&when, I_MINUTE,
	      (int) (dstamount (&when, 1) / MINUTE));
	}
    }

static void incrby (timebuf, code, amount)	/* Increment a time */
    reg1 struct tm *		timebuf;	/* Buffer to increment */
    reg2 int			code;		/* Code for increment unit */
    reg3 int			amount;		/* Amount to increment by */
    {

    if (code == I_WEEK)				/* Convert weeks to days */
	{
	code = I_DAY;
	amount *= 7;
	}
    switch (code)
	{
	case I_MINUTE:
	    timebuf->tm_min += amount;
	    if (timebuf->tm_min >= 0  &&  timebuf->tm_min < 60)
		break;
	    amount = timebuf->tm_min / 60;
	    timebuf->tm_min %= 60;
	    if (timebuf->tm_min < 0)
		{
		timebuf->tm_min += 60;
		amount--;
		}
	    /* FALLTHRU */
	case I_HOUR:
	    timebuf->tm_hour += amount;
	    if (timebuf->tm_hour >= 0  &&  timebuf->tm_hour < 24)
		break;
	    amount = timebuf->tm_hour / 24;
	    timebuf->tm_hour %= 24;
	    if (timebuf->tm_hour < 0)
		{
		timebuf->tm_hour += 24;
		amount--;
		}
	    /* FALLTHRU */
	case I_DAY:
	    timebuf->tm_mday += amount;
	    /*
	     * Now is the painful part.  We have to convert days into a
	     * day/month pair (it's really not all THAT bad).
	     */
	    while (timebuf->tm_mday < 0
	      ||  timebuf->tm_mday > mlens[timebuf->tm_mon]
	        + leapadj (timebuf->tm_mon, timebuf->tm_year))
		{
		if (timebuf->tm_mday < 0)
		    {
		    timebuf->tm_mon--;
		    if (timebuf->tm_mon < 0)
			{
			timebuf->tm_mon += 12;
			timebuf->tm_year--;
			}
		    timebuf->tm_mday +=
		      mlens[timebuf->tm_mon]
			+ leapadj (timebuf->tm_mon, timebuf->tm_year);
		    }
		else
		    {
		    timebuf->tm_mday -=
		      mlens[timebuf->tm_mon]
			+ leapadj (timebuf->tm_mon, timebuf->tm_year);
		    timebuf->tm_mon++;
		    if (timebuf->tm_mon >= 12)
			{
			timebuf->tm_mon -= 12;
			timebuf->tm_year++;
			}
		    }
		}
	    break;
	case I_MONTH:
	    timebuf->tm_mon += amount;
	    if (timebuf->tm_mon >= 0  &&  timebuf->tm_mon < 12)
		break;
	    amount = timebuf->tm_mon / 12;
	    timebuf->tm_mon %= 12;
	    if (timebuf->tm_mon < 0)
		{
		timebuf->tm_mon += 12;
		amount--;
		}
	    /* FALLTHRU */
	case I_YEAR:
	    timebuf->tm_year += amount;
	    break;
	}
    }

static int addjob ()			/* Add an at job */
    {
    reg2 int		ch;		/* Character being copied */
    char		dname[BUFSIZ];	/* Buffer for directory name */
    reg1 FILE *		outfile;	/* File we are writing to */

    outfile = popen ("pwd", "r");	/* Find out our current directory */
    (void) strcpy (dname, "/tmp");	/* Default current directory */
    if (outfile == NULL)
	{
	if (getenv ("HOME") != NULL)
	    {
	    (void) strncpy (dname, getenv ("HOME"), sizeof dname - 1);
	    dname[sizeof dname - 1] = '\0';
	    }
	}
    else
	{
	ch = fread (dname, 1, BUFSIZ, outfile);
	if (ch > 0  &&  dname[ch - 1] == '\n')
	    ch--;
	if (ch < 0)
	    ch = 0;
	dname[ch] = '\0';
	(void) pclose (outfile);
	}
    setlibdir ();
    (void) chdir (ATJOBDIR);

    /* If necessary, convert to 4-digit year */
    if (when.tm_year < 1900)
	    when.tm_year += 1900;

    /*
     * I'd love to use sprintf, but the syntax for forcing zero-fill varies
     * from system to system.  It's not too hard this way.
     *
     * For historical reasons having to do with 14-character
     * filenames, there are some at jobs out there that have 2-digit
     * years.  To get around the year-2000 problem, we need to flag
     * more modern jobs as having a 4-digit year.  This is done by
     * making the first character be non-numeric: the letter "a".
     *
     * To increase the probability of getting a unique string, we put
     * the seconds field from the current time into the when string.
     */
    (void) strcpy (whenstr, WHEN_PROTOTYPE); /* Re-init whenstr */
#ifdef LONGNAMES
    whenstr[WHEN_YEAR + 0] = (when.tm_year / 1000) % 10 + '0';
    whenstr[WHEN_YEAR + 1] = (when.tm_year / 100) % 10 + '0';
    whenstr[WHEN_YEAR + 2] = (when.tm_year / 10) % 10 + '0';
    whenstr[WHEN_YEAR + 3] = when.tm_year % 10 + '0';
#else /* LONGNAMES */
    whenstr[WHEN_YEAR + 0] = (when.tm_year / 10) % 10 + '0';
    whenstr[WHEN_YEAR + 1] = when.tm_year % 10 + '0';
#endif /* LONGNAMES */
    /*
     * We will make months 1-indexed so the user can read the file name.
     */
    when.tm_mon++;
    whenstr[WHEN_MONTH + 0] = when.tm_mon / 10 + '0';
    whenstr[WHEN_MONTH + 1] = when.tm_mon % 10 + '0';
    whenstr[WHEN_DAY] = when.tm_mday / 10 + '0';
    whenstr[WHEN_DAY + 1] = when.tm_mday % 10 + '0';
    whenstr[WHEN_HOUR + 0] = when.tm_hour / 10 + '0';
    whenstr[WHEN_HOUR + 1] = when.tm_hour % 10 + '0';
    whenstr[WHEN_MINUTE + 0] = when.tm_min / 10 + '0';
    whenstr[WHEN_MINUTE + 1] = when.tm_min % 10 + '0';
    whenstr[WHEN_ID + 0] = now->tm_sec / 10 + '0';
    whenstr[WHEN_ID + 1] = now->tm_sec % 10 + '0';

#ifdef LOC_CRON
#ifdef LOC_ATLOC
    if (where) {
	    /* start all over, a minor performance hit. */
	    (void) strcpy (whenstr, ATLOC_PROTOTYPE);
    };
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */

    /*
     * Catch signals so we can unlink the temp file
     */
    if (signal (SIGHUP, SIG_IGN) != SIG_IGN)
	(void) signal (SIGHUP, catchsigs);
    if (signal (SIGINT, SIG_IGN) != SIG_IGN)
	(void) signal (SIGINT, catchsigs);
#ifndef DEBUG
    if (signal (SIGILL, SIG_IGN) != SIG_IGN)
	(void) signal (SIGILL, catchsigs);
    if (signal (SIGFPE, SIG_IGN) != SIG_IGN)
	(void) signal (SIGFPE, catchsigs);
    if (signal (SIGBUS, SIG_IGN) != SIG_IGN)
	(void) signal (SIGBUS, catchsigs);
    if (signal (SIGSEGV, SIG_IGN) != SIG_IGN)
	(void) signal (SIGSEGV, catchsigs);
#endif
    if (signal (SIGTERM, SIG_IGN) != SIG_IGN)
	(void) signal (SIGTERM, catchsigs);
    /*
     * A comment is in order here about whenstr and uniqueness.  Since
     * 'tempfile' does not engage in any locking, a malicious user
     * could potentially generate a duplicate 'at' job name.  To get
     * the maximum possible uniqueness, we put the process ID and the
     * current seconds reading into the filename.  Since the process
     * ID is guaranteed by the kernel to be unique, this is sufficient
     * to prevent this attack unless the malicious user can regenerate
     * the process ID within the same second.  This can only be done
     * on a very fast modern machine, and under such systems, the
     * O_EXCL option is supported so that we can be guaranteed not to
     * overwrite an existing file.
     */
#ifdef HASFCNTL
    while (1)
	{
	/* Make sure whenstr has the trailing X's needed by tempfile */
#ifdef LOC_CRON
#ifdef LOC_ATLOC
	if (where)
	    (void) strcpy (&whenstr[ATLOC_X_TRAILER], ATLOC_TRAILER_PROTOTYPE);
	else (void) strcpy (&whenstr[WHEN_X_TRAILER], WHEN_TRAILER_PROTOTYPE);
#else /* ! LOC_ATLOC */
	(void) strcpy (&whenstr[WHEN_X_TRAILER], WHEN_TRAILER_PROTOTYPE);
#endif /* LOC_ATLOC */
#else /* ! LOC_CRON */
	(void) strcpy (&whenstr[WHEN_X_TRAILER], WHEN_TRAILER_PROTOTYPE);
#endif /* LOC_CRON */
	(void) tempfile (whenstr);		/* Make a unique file name */
	/*
	 * The options in the following open cause it to behave like 'creat',
	 * except it will fail if the file already exists (O_EXCL).
	 */
	ch = open (whenstr,
	  O_WRONLY | O_CREAT | O_TRUNC | O_EXCL,
	  (ATMODE | MUSTHAVE) & ~CANTHAVE);
	if (ch < 0)
	    {
	    if (errno != EEXIST)		/* File exists */
		{
		(void) chdir (LIBDIR);
		(void) fprintf (stderr, "Can't create job file named %s: %s\n",
		  whenstr, errmsg (0));
		return 1;
		}
	    }
	else					/* Open succeeded */
	    {
	    if ((outfile = fdopen (ch, "w")) != NULL)
		break;
	    else
		{
		(void) chdir (LIBDIR);
		(void) fprintf (stderr,
		  "Can't fdopen job file named %s: %s\n",
		  whenstr, errmsg (0));
		return 1;
		}
	    }
	}
#else
    (void) tempfile (whenstr);			/* Make a unique file name */
    if ((outfile = fopen (whenstr, "w")) == NULL)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr, "Can't create job file named %s: %s\n",
	  whenstr, errmsg (0));
	return 1;
	}
#endif
    /*
     * Make the job file's ownership and modes correct.
     */
    if (chown (whenstr, userpwent->pw_uid, userpwent->pw_gid) < 0)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr, "at:  couldn't change owner of %s: %s\n",
	  whenstr, errmsg (0));
	(void) fclose (outfile);
	(void) unlink (whenstr);
	return 1;
	}
    if (chmod (whenstr, (ATMODE | MUSTHAVE) & ~CANTHAVE) < 0)
	{
	(void) chdir (LIBDIR);
	(void) fprintf (stderr, "at:  couldn't change mode of %s: %s\n",
	  whenstr, errmsg (0));
	(void) fclose (outfile);
	(void) unlink (whenstr);
	return 1;
	}
    /*
     * We've created the output file in the spool directory.  Put out the
     * file prefix, followed by copying the commands to the output file.
     * Finish with the file suffix.
     */
    fprefix (outfile, dname);
    while ((ch = getchar ()) != EOF)
	(void) putc (ch, outfile);
    fsuffix (outfile);
    (void) fclose (outfile);
    (void) chdir (LIBDIR);
    /*
     * Now tell the cron daemon to schedule the file.  We ignore error no. 1
     * because it's not considered a failure if the daemon isn't running.
     */
    if (sendcommand ('A', whenstr, 1) == SEND_ERROR)
	return 1;
    else
	return 0;
    }

static void fprefix (file, dname) /* Put out the at-job file prefix */
    reg1 FILE *		file;	/* File to write to */
    char *		dname;	/* Directory the job should run in */
    {
    reg3 char **	envp;	/* List of environment variables */
    reg2 char *		envvar;	/* Ptr into current environment variable */
    int			mask;	/* User's file creation mask */
    reg4 char *		usershell; /* Name of the user's shell */

#ifdef LOC_CRON
#ifdef LOC_ATLOC
    /*
     * atloc jobs have type-'b' prefixes.
     * note that tm is cron-specific at this point, not what
     * libc wants (i.e., year is >1900, month is 1-based, etc.
     * tm is already incremented
     */
    (void) fprintf(file, "#Y%d\n#m%d\n#d%d\n#H%d\n#M%d\n#@%s\n#.\n",
		   when.tm_year, when.tm_mon, when.tm_mday, when.tm_hour, when.tm_min, where);
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */
    /*
     * Strip off "/tmp_mnt" if necessary.
     */
    if (strncmp (dname, AUTOMOUNTPOINT, sizeof AUTOMOUNTPOINT - 1) == 0)
	dname += sizeof AUTOMOUNTPOINT - 1;
    (void) fprintf (file, "cd %s\n", dname);
    mask = umask (0);
    (void) umask (mask);
    (void) fprintf (file, "umask %o\n", mask);
    for (envp = environ;  *envp != NULL;  envp++)
	{
	for (envvar = *envp;  *envvar != '\0'  &&  *envvar != '=';  envvar++)
	    (void) putc (*envvar, file);
	(void) putc ('=', file);
	(void) putc ('\'', file);
	if (*envvar != '\0')
	    {
	    for (envvar++;  *envvar != '\0';  envvar++)
		{
		switch (*envvar)
		    {
		    case '\'':
			(void) fprintf (file, "'\"'\"'");
			break;
		    case '\\':
			(void) fprintf (file, "\\\\");
			break;
		    default:
			(void) putc (*envvar, file);
			break;
		    }
		}
	    }
	(void) fprintf (file, "'\nexport ");
	for (envvar = *envp;  *envvar != '\0'  &&  *envvar != '=';  envvar++)
	    (void) putc (*envvar, file);
	(void) putc ('\n', file);
	}
    if ((usershell = getenv ("SHELL")) == NULL)
	usershell = ATSHELL;	/* Just use the same shell */
    (void) fprintf (file, "exec %s << '%s'\n", usershell, UNIQUESTR);
    }

static void fsuffix (file)	/* Put out the at-job file suffix */
    FILE *		file;	/* File to write to */
    {
    (void) fprintf (file, "%s\n", UNIQUESTR);
    }

static SIG_TYPE catchsigs (sig)	/* Catch unfriendly signals and clean up */
    int			sig;	/* Signal we received */
    {
    (void) unlink (whenstr);	/* Get rid our temp file */
    (void) kill (getpid (), sig); /* Re-deliver signal to kill ourselves */
    /* NOTREACHED */
    }

/*
 * extract the at stuff from the file for type-b files
 */
#ifdef LOC_CRON
#ifdef LOC_ATLOC
static void atloc_decode_at_file (name, year, month, day, hour, minute, locp)
    reg2 char *			name;	/* Name of the file holding the job */
    reg1 int *			year;	/* Year job should run */
    int *			month;	/* Month job should run */
    int *			day;	/* Day job should run */
    int *			hour;	/* Hour job should run */
    int *			minute;	/* Minute job should run */
    char **			locp; /* location, only for ATLOC */
    {
    FILE *f;
/* xxx */
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif /* ! MAXPATHLEN */
    char buf[MAXPATHLEN];
    char *bufend;
    char kind;
    int n;

    /*
     * There's a possible buffer overflow here,
     * but since both ATJOBDIR and name
     * are sysadmin or xcron (not user) generated
     * we trust them.
     */
    sprintf(buf, "%s/%s", ATJOBDIR, name);
    if ((f = fopen(buf, "r")) == NULL) {
	    fprintf(stderr, "atloc_decode_at_file: internal problem, cannot read file %s.\n", name);
	    return;
    };
    /* Must have defaults for everything since something
     * may be missing.  Pick a date in the past so we can
     * handle edge-triggered jobs.
     */
    *year = ATLOC_NOYEAR;
    *month = 1;
    *day = 1;
    *hour = 0;
    *minute = 0;
    if (locp)
	    *locp = NULL;
    while (fgets(buf, sizeof buf, f)) {
	    if (buf[0] != '#') /* everything hidden in shell comments */
		    break;
	    if (buf[1] == '.') /* end of attributes */
		    break;
	    kind = buf[1];
	    /* perl's chomp */
	    bufend = &buf[strlen(buf) - 1];
	    assert(*bufend == '\n');
	    *bufend = 0;
	    n = atoi(&buf[2]); /* just in case */
	    switch (kind) {
	    case 'Y': /* 4-digit year */
		    *year = n;
		    break;
	    case 'm': /* month (1-based) */
		    *month = n;
		    break;
	    case 'd': /* day of month (1-based) */
		    *day = n;
		    break;
	    case 'H': /* 24 hour */
		    *hour = n;
		    break;
	    case 'M': /* minute */
		    *minute = n;
		    break;
	    case '@': /* location */
		    if (!locp)  /* can pass null locp */
			    break;
		    if (*locp)
			    free(*locp);
		    *locp = strdup(&buf[2]);
		    break;
	    default:
		    fprintf(stderr, "atloc_decode_at_file: internal problem, unknown kind %c on line %s.\n", kind, buf);
		    break;
	    };
    };
    fclose(f);
    }
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */

static int	listjobs ()		/* List at jobs belonging to user */
    {
    char		buf[512];	/* Buffer for command lines */
    reg3 DIR *		dir;		/* Directory we are reading */
    reg1 struct direct * entry;		/* Current directory entry */
    char *		getresult;	/* Result of fgets */
    reg4 FILE *		jobfile;	/* File for current job */
    reg5 struct jobname * jobnames = NULL; /* Sorted list of relevant jobs */
    reg2 int		len;		/* Length of data in buffer */
    reg7 int		maxjobnames = 0; /* Size of jobnames array */
    reg6 int		njobnames = 0;	/* No. of job names currently known */
    struct passwd *	pwent;		/* Password entry for job's UID */
    struct stat		statbuf;	/* Status of current entry */
    int			useruid;	/* Uid of the invoking user */
    int			year;		/* Year the job will execute */

    useruid = getuid ();		/* Get invoker's UID */
    setlibdir ();
    dir = opendir (ATJOBDIR);
    if (dir == NULL)
	{
	(void) fprintf (stderr, "at: Couldn't open %s: %s\n", ATJOBDIR,
	  errmsg (0));
	return 1;
	}
    (void) chdir (ATJOBDIR);
    while ((entry = readdir (dir)) != NULL)
	{
	if (strcmp (entry->d_name, ".") != 0
	  &&  strcmp (entry->d_name, "..") != 0)
	    {
	    if (stat (entry->d_name, &statbuf) < 0)
		(void) fprintf (stderr, "at:  Couldn't stat %s: %s\n",
		  entry->d_name, errmsg (0));
	    else if (useruid == 0  ||  statbuf.st_uid == useruid)
		{
		if (njobnames >= maxjobnames)
		    {
		    maxjobnames += LISTJOB_INCR;
		    if (jobnames == NULL)
			jobnames = (struct jobname *) malloc (
			  maxjobnames * sizeof (struct jobname));
		    else
			jobnames = (struct jobname *) realloc (
			  (char *) jobnames,
			  maxjobnames * sizeof (struct jobname));
		    if (jobnames == NULL)
			{
			(void) fprintf (stderr,
			  "Unable to allocate %d bytes for at job list\n",
			  maxjobnames * sizeof (struct jobname));
			return 1;
			}
		    }
		jobnames[njobnames].name = malloc (strlen (entry->d_name) + 1);
		if (jobnames[njobnames].name == NULL)
		    {
		    (void) fprintf (stderr,
		      "Unable to allocate %d bytes for name of job '%s'\n",
		      strlen (entry->d_name) + 1, entry->d_name);
		    return 1;
		    }
		(void) strcpy (jobnames[njobnames].name, entry->d_name);
		jobnames[njobnames].uid = statbuf.st_uid;
		njobnames++;
		}
	    }
	}
    closedir (dir);
    qsort ((char *) jobnames, njobnames, sizeof jobnames[0], comparejobs);
    maxjobnames = njobnames;		/* This is now actual number of jobs */
    for (njobnames = 0;  njobnames < maxjobnames;  njobnames++)
	{
	if (useruid == 0)	/* Root needs to see user name */
	    {
	    pwent = getpwuid (jobnames[njobnames].uid);
	    if (pwent == NULL)
		(void) printf ("%d\t", jobnames[njobnames].uid);
	    else
		(void) printf ("%s\t", pwent->pw_name);
	    }
#ifdef LONGNAMES
	if (jobnames[njobnames].name[WHEN_FLAG] == WHEN_FLAG_VALUE)
	    {
	    (void) printf ("%4.4s %2.2s/%2.2s at %2.2s:%2.2s ID %s",
	      jobnames[njobnames].name + WHEN_YEAR,
	      jobnames[njobnames].name + WHEN_MONTH,
	      jobnames[njobnames].name + WHEN_DAY,
	      jobnames[njobnames].name + WHEN_HOUR,
	      jobnames[njobnames].name + WHEN_MINUTE,
	      jobnames[njobnames].name + WHEN_ID);
	    }
#ifdef LOC_CRON
#ifdef LOC_ATLOC
	else if (jobnames[njobnames].name[WHEN_FLAG] == ATLOC_FLAG_VALUE) {
		int year, month, day, hour, minute;
		char *loc;
		atloc_decode_at_file(jobnames[njobnames].name, &year, &month, &day, &hour, &minute, &loc);
		if (!loc)
			loc = strdup("");
		(void) printf ("%4d %2d/%2d at %02d:%02d @%s ID %s",
			   year, month, day, hour, minute, loc, jobnames[njobnames].name + ATLOC_X_TRAILER);
		if (loc)
			free(loc);
		loc = NULL;

	}
#endif /* LOC_ATLOC */
#endif /* LOC_CRON */
	else
	    {
	    (void) sscanf (&jobnames[njobnames].name[WHEN_OLD_YEAR], "%2d",
	      &year);
	    if (year < Y2K_THRESHOLD)
		year += Y2K_2000;
	    else
		year += Y2K_1900;
	    (void) printf ("%4.4d %2.2s/%2.2s at %2.2s:%2.2s ID %s",
	      year,
	      jobnames[njobnames].name + WHEN_OLD_MONTH,
	      jobnames[njobnames].name + WHEN_OLD_DAY,
	      jobnames[njobnames].name + WHEN_OLD_HOUR,
	      jobnames[njobnames].name + WHEN_OLD_MINUTE,
	      jobnames[njobnames].name + WHEN_OLD_ID);
	    }
#else /* LONGNAMES */
	(void) sscanf (&jobnames[njobnames].name[WHEN_YEAR], "%2d", &year);
	if (year < Y2K_THRESHOLD)
	    year += Y2K_2000;
	else
	    year += Y2K_1900;
	(void) printf ("%4.4d %2.2s/%2.2s at %2.2s:%2.2s ID %s",
	  year,
	  jobnames[njobnames].name + WHEN_MONTH,
	  jobnames[njobnames].name + WHEN_DAY,
	  jobnames[njobnames].name + WHEN_HOUR,
	  jobnames[njobnames].name + WHEN_MINUTE,
	  jobnames[njobnames].name + WHEN_ID);
#endif /* LONGNAMES */
	if (!verbose  &&  useruid != jobnames[njobnames].uid)
	    (void) printf ("\n");
	else
	    {
	    jobfile = fopen (jobnames[njobnames].name, "r");
	    if (jobfile == NULL)
		{
		(void) printf ("\n");
		(void) fprintf (stderr, "at: Couldn't open %s: %s\n",
		  jobnames[njobnames].name, errmsg (0));
		}
	    else
		{
		/*
		 * Look through the file trying to find the line
		 * $shell << 'UNIQUESTR'.  If we find it, print
		 * out the following lines until we reach UNIQUESTR.
		 * These are presumably the commands in the job.
		 */
		while ((getresult = fgets (buf, sizeof buf, jobfile))
		  != NULL)
		    {
		    len = strlen (buf);
		    if (buf[len - 1] != '\n')
			{
			while ((len = getc (jobfile)) != '\n'
			  &&  len != EOF)
			    ;
			}
		    else if (len > sizeof UNIQUESTR + 2
		      &&  strncmp (&buf[len - sizeof UNIQUESTR - 1],
			UNIQUESTR, sizeof UNIQUESTR - 1) == 0)
			{
			getresult = NULL;
			(void) printf ("\n");
			while (fgets (buf, sizeof buf, jobfile) != NULL)
			    {
			    if (strncmp (buf, UNIQUESTR,
				sizeof UNIQUESTR - 1)
			      == 0
			      &&  buf[sizeof UNIQUESTR - 1] == '\n')
				break;	/* End of commands */
			    (void) printf ("  %s", buf);
			    getresult = buf; /* Mark that cmds were found */
			    }
			break;
			}
		    }
		if (getresult == NULL)
		    (void) printf ("\t(No command found)\n");
		(void) fclose (jobfile);
		}
	    }
	free (jobnames[njobnames].name);
	jobnames[njobnames].name = NULL;
	}
    free ((char *) jobnames);
    return 0;
    }

static int	comparejobs (job1, job2) /* Compare job names for sorting */
    struct jobname *	job1;		/* Name of first job */
    struct jobname *	job2;		/* Name of second job */
    {
    int			result;		/* Result of strcmp */
    int			year1;		/* Year of job 1 */
    int			year2;		/* Year of job 2 */

#ifdef LONGNAMES
    if (job1->name[WHEN_FLAG] == WHEN_FLAG_VALUE
      &&  job2->name[WHEN_FLAG] == WHEN_FLAG_VALUE)
	{
	year1 = year2 = 0;
	result = strncmp (job1->name, job2->name, WHEN_ID);
	}
    else if (job1->name[WHEN_FLAG] == WHEN_FLAG_VALUE)
	{
	(void) sscanf (&job1->name[WHEN_YEAR], "%4d", &year1);
	(void) sscanf (&job2->name[WHEN_OLD_YEAR], "%2d", &year2);
	if (year2 < Y2K_THRESHOLD)
	    year2 += Y2K_2000;
	else
	    year2 += Y2K_1900;
	result = strncmp (&job1->name[WHEN_YEAR + 2],
	  &job2->name[WHEN_OLD_YEAR], WHEN_OLD_ID - WHEN_OLD_YEAR);
	}
    else if (job2->name[WHEN_FLAG] == WHEN_FLAG_VALUE)
	{
	(void) sscanf (&job1->name[WHEN_OLD_YEAR], "%2d", &year1);
	(void) sscanf (&job2->name[WHEN_YEAR], "%4d", &year2);
	if (year1 < Y2K_THRESHOLD)
	    year1 += Y2K_2000;
	else
	    year1 += Y2K_1900;
	result = strncmp (&job1->name[WHEN_OLD_YEAR],
	  &job2->name[WHEN_YEAR + 2], WHEN_OLD_ID - WHEN_OLD_YEAR);
	}
    else
	{
	(void) sscanf (&job1->name[WHEN_OLD_YEAR], "%2d", &year1);
	(void) sscanf (&job2->name[WHEN_OLD_YEAR], "%2d", &year2);
	if (year1 < Y2K_THRESHOLD)
	    year1 += Y2K_2000;
	else
	    year1 += Y2K_1900;
	if (year2 < Y2K_THRESHOLD)
	    year2 += Y2K_2000;
	else
	    year2 += Y2K_1900;
	result =
	  strncmp (&job1->name[WHEN_OLD_MONTH], &job2->name[WHEN_OLD_MONTH],
	    WHEN_ID - WHEN_OLD_MONTH);
	}
#else /* LONGNAMES */
    (void) sscanf (&job1->name[WHEN_YEAR], "%2d", &year1);
    (void) sscanf (&job2->name[WHEN_YEAR], "%2d", &year2);
    if (year1 < Y2K_THRESHOLD)
	year1 += Y2K_2000;
    else
	year1 += Y2K_1900;
    if (year2 < Y2K_THRESHOLD)
	year2 += Y2K_2000;
    else
	year2 += Y2K_1900;
    result = strncmp (&job1->name[WHEN_MONTH], &job2->name[WHEN_MONTH],
      WHEN_ID - WHEN_MONTH);
#endif /* LONGNAMES */
    if (year1 != year2)
	return year1 - year2;
    if (result == 0)
	{
	result = job1->uid - job2->uid;	/* If times equal, sort on UID's */
	if (result == 0)
	    {
	    /* Equal UID's, use job ID as tiebreaker */
#ifdef LONGNAMES
	    if (job1->name[WHEN_FLAG] == job2->name[WHEN_FLAG])
		return strcmp (job1->name, job2->name);
	    else if (job1->name[WHEN_FLAG] == WHEN_FLAG_VALUE)
		return strcmp (&job1->name[WHEN_X_TRAILER],
		  &job2->name[WHEN_OLD_X_TRAILER]);
	    else
		return strcmp (&job1->name[WHEN_OLD_X_TRAILER],
		  &job2->name[WHEN_X_TRAILER]);
#else /* LONGNAMES */
	    return strcmp (job1->name, job2->name);
#endif /* LONGNAMES */
	    }
	}
    return result;
    }

static int	remjob (id, anytime)	/* Remove an at job from the queue */
    char *		id;		/* ID of the job to be removed */
    int			anytime;	/* NZ if any time is acceptable */
    {
    reg2 DIR *		dir = NULL;	/* Directory we are reading */
    reg1 struct direct * entry;		/* Current directory entry */
    reg3 int		found = 0;	/* NZ if already found an entry */
#ifdef LONGNAMES
    char		oldwhenstr[sizeof WHEN_OLD_PROTOTYPE];
					/* Old-format version of whenstr */
#endif /* LONGNAMES */
    struct stat		statbuf;	/* Status of current entry */
    int			useruid;	/* Uid of the invoking user */

    useruid = getuid ();		/* Get invoker's UID */
    setlibdir ();
#if MAXNAMLEN < 20
    if (checkwhen (1))
	return 1;			/* Illegal date */
#else /* MAXNAMLEN */
    if (anytime)
	{
	dir = opendir (ATJOBDIR);
	if (dir == NULL)
	    {
	    (void) fprintf (stderr, "at: Couldn't open %s: %s\n", ATJOBDIR,
	      errmsg (0));
	    return 1;
	    }
	(void) chdir (ATJOBDIR);
	while ((entry = readdir (dir)) != NULL)
	    {
	    if (strcmp (entry->d_name, ".") != 0
	      &&  strcmp (entry->d_name, "..") != 0)
		{
		if (stat (entry->d_name, &statbuf) < 0)
		    (void) fprintf (stderr, "at:  Couldn't stat %s: %s\n",
		      entry->d_name, errmsg (0));
		else
		    {
#ifdef LONGNAMES
		    if (strcmp (entry->d_name
			  + (entry->d_name[WHEN_FLAG] == WHEN_FLAG_VALUE
			    ? WHEN_ID : WHEN_OLD_ID),
			  id)
			== 0
		      &&  (useruid == 0  ||  statbuf.st_uid == useruid))
#else /* LONGNAMES */
		    if (strcmp (entry->d_name + WHEN_ID, id) == 0
		      &&  (useruid == 0  ||  statbuf.st_uid == useruid))
#endif /* LONGNAMES */
			{
			/*
			 * This entry matches the ID requested, and
			 * the user is authorized to delete it.  If
			 * found is nonzero, there is more than one
			 * match, and we must abort.  Otherwise, we
			 * save the important information in 'when'.
			 * (We can't just save a pointer to the
			 * directory entry, because the directory
			 * library might clobber it.  This means we
			 * have to read the directory twice, but
			 * that's not too expensive to accept.)
			 */
			if (found)
			    {
			    (void) fprintf (stderr,
			      "at:  ID '%s' is ambiguous\n",
			      id);
			    return 1;
			    }
			found = 1;
#ifdef LONGNAMES
			if (entry->d_name[WHEN_FLAG] == WHEN_FLAG_VALUE)
			    (void) sscanf (&entry->d_name[WHEN_YEAR],
			      "%4d%2d%2d%2d%2d",
			      &when.tm_year, &when.tm_mon, &when.tm_mday,
			      &when.tm_hour, &when.tm_min);
			else
			    {
			    (void) sscanf (&entry->d_name[WHEN_OLD_YEAR],
			      "%2d%2d%2d%2d%2d",
			      &when.tm_year, &when.tm_mon, &when.tm_mday,
			      &when.tm_hour, &when.tm_min);
			    if (when.tm_year < Y2K_THRESHOLD)
				when.tm_year += Y2K_2000;
			    else
				when.tm_year += Y2K_1900;
			    }
#else /* LONGNAMES */
			(void) sscanf (&entry->d_name[WHEN_YEAR],
			  "%2d%2d%2d%2d%2d",
			  &when.tm_year, &when.tm_mon, &when.tm_mday,
			  &when.tm_hour, &when.tm_min);
			 if (when.tm_year < Y2K_THRESHOLD)
			     when.tm_year += Y2K_2000;
			 else
			     when.tm_year += Y2K_1900;
#endif /* LONGNAMES */
			when.tm_mon--;
			}
		    }
		}
	    }
	if (!found)
	    {
	    (void) fprintf (stderr, "at:  Job '%s' not found\n", id);
	    return 1;
	    }
	if (checkwhen (1))
	    {
	    (void) fprintf (stderr,
	      "at:  The above error was due to a bug!\n");
	    return 1;
	    }
	}
    else
	{
	if (checkwhen (1))
	    return 1;			/* Illegal date */
	}
#endif /* MAXNAMLEN */
    fixwhen ();				/* Fill in unspecified fields */
    when.tm_year += 1900;		/* Convert to 4-digit year */
    /*
     * I'd love to use sprintf, but the syntax for forcing zero-fill varies
     * from system to system.  It's not too hard this way.
     */
    (void) strcpy (whenstr, WHEN_PROTOTYPE); /* Re-init whenstr */
#ifdef LONGNAMES
    whenstr[WHEN_YEAR + 0] = (when.tm_year / 1000) % 10 + '0';
    whenstr[WHEN_YEAR + 1] = (when.tm_year / 100) % 10 + '0';
    whenstr[WHEN_YEAR + 2] = (when.tm_year / 10) % 10 + '0';
    whenstr[WHEN_YEAR + 3] = when.tm_year % 10 + '0';
#else /* LONGNAMES */
    whenstr[WHEN_YEAR + 0] = (when.tm_year / 10) % 10 + '0';
    whenstr[WHEN_YEAR + 1] = when.tm_year % 10 + '0';
#endif /* LONGNAMES */
    /*
     * We make months 1-indexed so the user can read the file name.
     */
    when.tm_mon++;
    whenstr[WHEN_MONTH + 0] = when.tm_mon / 10 + '0';
    whenstr[WHEN_MONTH + 1] = when.tm_mon % 10 + '0';
    whenstr[WHEN_DAY + 0] = when.tm_mday / 10 + '0';
    whenstr[WHEN_DAY + 1] = when.tm_mday % 10 + '0';
    whenstr[WHEN_HOUR + 0] = when.tm_hour / 10 + '0';
    whenstr[WHEN_HOUR + 1] = when.tm_hour % 10 + '0';
    whenstr[WHEN_MINUTE + 0] = when.tm_min / 10 + '0';
    whenstr[WHEN_MINUTE + 1] = when.tm_min % 10 + '0';
    (void) strncpy (whenstr + WHEN_ID, id, sizeof whenstr - WHEN_ID - 1);
#ifdef LONGNAMES
    (void) strcpy (oldwhenstr, WHEN_OLD_PROTOTYPE); /* Init oldwhenstr */
    oldwhenstr[WHEN_OLD_YEAR + 0] = (when.tm_year / 10) % 10 + '0';
    oldwhenstr[WHEN_OLD_YEAR + 1] = when.tm_year % 10 + '0';
    oldwhenstr[WHEN_OLD_MONTH + 0] = when.tm_mon / 10 + '0';
    oldwhenstr[WHEN_OLD_MONTH + 1] = when.tm_mon % 10 + '0';
    oldwhenstr[WHEN_OLD_DAY + 0] = when.tm_mday / 10 + '0';
    oldwhenstr[WHEN_OLD_DAY + 1] = when.tm_mday % 10 + '0';
    oldwhenstr[WHEN_OLD_HOUR + 0] = when.tm_hour / 10 + '0';
    oldwhenstr[WHEN_OLD_HOUR + 1] = when.tm_hour % 10 + '0';
    oldwhenstr[WHEN_OLD_MINUTE + 0] = when.tm_min / 10 + '0';
    oldwhenstr[WHEN_OLD_MINUTE + 1] = when.tm_min % 10 + '0';
    (void) strncpy (oldwhenstr + WHEN_OLD_ID, id,
      sizeof oldwhenstr - WHEN_OLD_ID - 1);
#endif /* LONGNAMES */
    if (dir == NULL)
	{
	dir = opendir (ATJOBDIR);
	if (dir == NULL)
	    {
	    (void) fprintf (stderr, "at: Couldn't open %s: %s\n", ATJOBDIR,
	      errmsg (0));
	    return 1;
	    }
	(void) chdir (ATJOBDIR);
	}
    else
	seekdir (dir, 0);
    /*
     * Read the at job directory looking for the specified job.
     *
     * NEEDSWORK: should allow ID to be unspecified if when is given.
     */
    while ((entry = readdir (dir)) != NULL)
	{
	if (strcmp (entry->d_name, ".") != 0
	  &&  strcmp (entry->d_name, "..") != 0)
	    {
	    if (stat (entry->d_name, &statbuf) < 0)
		(void) fprintf (stderr, "at:  Couldn't stat %s: %s\n",
		  entry->d_name, errmsg (0));
#ifdef LONGNAMES
	    else if (strncmp (entry->d_name, whenstr, MAXNAMLEN - 1) == 0
	      ||  strncmp (entry->d_name, oldwhenstr, MAXNAMLEN - 1) == 0)
#else /* LONGNAMES */
	    else if (strncmp (entry->d_name, whenstr, MAXNAMLEN - 1) == 0)
#endif /* LONGNAMES */
		{
		if (useruid == 0  ||  statbuf.st_uid == useruid)
		    {
		    /*
		     * We've found our victim!
		     */
		    (void) unlink (entry->d_name);
		    /*
		     * Now tell the cron daemon to trash the file.
		     */
		    (void) chdir (LIBDIR);
		    if (sendcommand ('d', entry->d_name, 0) == SEND_ERROR)
			return 1;
		    else
			return 0;
		    }
		else
		    {
		    (void) fprintf (stderr, "at:  You don't own that job\n");
		    closedir (dir);
		    return 1;
		    }
		}
	    }
	}
	closedir (dir);
	(void) fprintf (stderr, "at:  Job '%s' not found\n", id);
	return 1;
    }
