#!/usr/bin/python

#
# base.py
# $Id: dbus_monitor.py 4551 2007-01-09 18:56:23Z johnh $
#
# Copyright (C) 2006 by John Heidemann <johnh@isi.edu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#

import sys
import os
import getopt
import re
import time

import gobject 
import dbus
if getattr(dbus, 'version', (0,0,0)) >= (0,41,0):
    import dbus.glib

    
def usage():
    print sys.stderr, """
usage: program

watches dbus and triggers lcron appropriately
(replaces location_monitor.pl)

options:
-D	run in daemon mode
-R	silently ignored (for compatibility with location_monitor)
    """
    sys.exit(1)

#
# xcron stuff:
#
XCRON_BINDIR = '/usr/local/lib/xcron/bin'
XCRON_LIBDIR = '/usr/local/lib/xcron/lib'
NETSTAT = "/bin/netstat";

#
# argument parsing
#
verbose = False


def save_pid():
    """write the pid to the system run dir"""
    pid_filename = "/var/run/dbus_monitor.pid"
    try:
        pid_f = open(pid_filename, 'w');
    except IOError:
        print "cannot write " + pid_filename
        sys.exit(1)
    pid_f.write(str(os.getpid()) + "\n");
    pid_f.close();
    return

def check_once():
    """check current network status"""
    if verbose:
        print "check_once at " . ctime();

    # get the network
    try:
        ns = os.popen(NETSTAT + " -rn", "r")
    except IOError:
        print "cannot run netstat"
        sys.exit(1);
    ns_lines = ns.readlines()
    ns.close
    # and our dictionary
    try:
        gw = open(XCRON_LIBDIR + "/events/gateways", "w", 0644)
    except IOError:
        print "cannot open " + XCRON_LIBDIR + "/events/gateways"
        sys.exit(1);
    gws = []
    for ln in ns_lines:
        f = ln.split()
        for addr in f[:2]:
            if not addr in ('0.0.0.0', 'Destination', 'Gateway', 'Kernel', 'IP'):
                gw.write(addr + "\n");
                gws.append(addr);
    gw.close()

    # now map to place
    gw_to_place = {}
    try:
        map = open(XCRON_LIBDIR + "/netstat_mapping", "r")
    except IOError:
        pass
    else:
        for ln in map.readlines():
            if re.match("\s*#", ln):  # skip comments
                continue  
            f = ln.split()
            if len(f) == 0:
                continue   # skip empty lines
            gw = f[0]
            gw_to_place[gw] = f[1:]
        map.close()

    # write new places
    try:
        pl = open(XCRON_LIBDIR + "/events/places", "w", 0644)
    except IOError:
        print "error: cannot write to " + XCRON_LIBDIR + "/events/places"
        sys.exit(1)
    for gw in gws:
        if gw_to_place.has_key(gw):
            for place in gw_to_place[gw]:
                pl.write(place + "\n")
                # write(place + "\n")
    pl.close()

    # bump xcron
    os.system(XCRON_BINDIR + "/tellxcron -L &")
    return

def network_change(udi):
    check_once()
    return


def check_loop():
    """attach to dbus and loop forever"""
    # add signals
    bus = dbus.SystemBus()
    for i in range(4):
        try:
            bus.add_signal_receiver(network_change,
                                    'DeviceNoLongerActive',
                                    'org.freedesktop.NetworkManager',
                                    'org.freedesktop.NetworkManager',
                                    '/org/freedesktop/NetworkManager')
        except dbus.DBusException:
            print "warning: dbus not up.  delaying and retrying"
            time.sleep(5)
            
    bus.add_signal_receiver(network_change,
                        'DeviceNowActive',
                        'org.freedesktop.NetworkManager',
                        'org.freedesktop.NetworkManager',
                        '/org/freedesktop/NetworkManager')

    # run forever
    mainloop = gobject.MainLoop()
    mainloop.run()
    return

def main():
    daemon = False
    try:
        opts, args = getopt.getopt(sys.argv[1:], "DRhv", ["daemon", "help", "verbose"])
    except getopt.GetoptError:
        # print help information and exit:
        usage()
    for o, a in opts:
        if o in ("-v", "--verbose"):
            verbose = True
        if o in ("-h", "--help"):
            usage()
        if o in ("-D", "--daemon"):
            daemon = True
    if len(sys.argv) == 0:
        usage()
    if len(sys.argv) == 1 and sys.argv[0] == '-?':
        usage()
    #
    # daemon mode
    if daemon:
        f = os.fork();
        if f == -1:
            sys.stderr.write("fork failed\n")
        if f != 0:
            # in parent => quit
            sys.exit(0)
        save_pid()
        pass
    check_once()
    check_loop()
    sys.exit(0);
        
if __name__ == "__main__":
    main()
