#ifndef lint
static char Rcs_Id[] =
    "$Header$";
#endif

/*
 * $Log$
 * Revision 2.7  1999/07/14 05:22:30  johnh
 * location awareness (all in LOC_CRON) and a few cleanups/bugfixes
 *
 * Revision 2.6  1998/12/17 23:10:00  geoff
 * If an at job has passed its run time, return now+1 so that we'll run
 * it right away.
 *
 * Revision 2.5  1998/07/30 20:42:26  geoff
 * Many changes, accumulated over far too much time: add a lot of
 * commentary, and improve the time-bumping code (I no longer remember
 * why).  Fix some problems with "at" jobs scheduled during the
 * daylight-savings changeovers.
 *
 * Revision 2.4  1995/05/11  05:05:42  geoff
 * Make stimebuf be a buffer, not a pointer, using structure copy to pick
 * up the results from localtime().
 *
 * Revision 2.3  86/11/29  19:52:34  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.2  86/04/13  13:32:56  geoff
 * Remove an unused external declaration.
 * 
 * Revision 2.1  86/03/28  13:42:05  geoff
 * Initial Revision.
 * 
 */

#include "cron.h"
#include <time.h>
#include <errno.h>
#include <stdio.h>

time_t		nextrun ();	/* Figure out when something runs next */
static int	nextset ();	/* Find next set bit */

/*
 * Leap-year calculation.  This will work for all dates in the future until
 * the year 2100, by which time I hope this software is thoroughly obsolete.
 * Also note that it works whether applied to a 4-digit year or a Unix
 * "year - 1900".
 */
#define leap(year)	(((year) & 3) == 0)

/*
 * Leap-year adjustment calculation.  Generates 1 if month is January or
 * February and the year is a leap year.  Note months are indexed from zero.
 */
#define leapadj(month, year)	((month) <= 1  &&  leap (year))

extern time_t	tm_to_time ();	/* Convert struct tm to time_t */
extern int	wkday ();	/* Calculate day of the week */
extern int	yearbase;	/* Base of year calculations */

time_t nextrun (stime, entp)	/* Figure out when something runs next */
    time_t		stime;	/* Starting time to work from */
    reg2 struct cronent * entp;	/* Crontab entry to be evaluated. */
    {
    enum {BUMPNONE, BUMPMINUTE, BUMPHOUR, BUMPDAY, BUMPMONTH, BUMPYEAR}
			bumpfield = BUMPNONE; /* What field to bump */
    static char		mlens[] =	/* Table of month lengths */
			    {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
    time_t		nexttime; /* Time to run the job next */
    struct tm		stimebuf; /* Broken-down value of stime */

#ifdef LOC_CRON
#ifdef LOC_MOVE
    if (entp->ce_location != NULL) {
	    /*
	     * Run edge-detected jobs immediately.
	     */
	    if (entp->ce_flags & EF_ONLY_ON_MOVE)
		    return time(NULL) + 1;
    };
#endif /* LOC_MOVE */
#endif /* LOC_CRON */

    stimebuf = *localtime (&stime);	/* Break stime into its components */
    /*
     * Loop until we find a valid time or run out of years.  The
     * looping is necessary because at the daylight-savings switch, we
     * might generate an invalid time, and if tm_to_time is based on
     * mktime, mktime will reject it and return -1.
     *
     * There are two exits from this loop, both via returns to the
     * caller.  The first is in the BUMPYEAR code, when we discover
     * that there are no more years left.  The second is near the end
     * of the loop, after we convert back to a time_t.  There, we
     * return if we find we've generated a valid time.
     */
    while (1)
	{
	stimebuf.tm_mday--;		/* Zero-index */
	stimebuf.tm_year -= (yearbase - 1900); /* Zero index */
	if (stimebuf.tm_year < 0)	/* Before this year? */
	    {				/* Set to Jan 1, yearbase */
	    stimebuf.tm_min = 0;
	    stimebuf.tm_hour = 0;
	    stimebuf.tm_mday = 0;
	    stimebuf.tm_mon = 0;
	    stimebuf.tm_year = 0;	/* This is equiv to "yearbase" */
	    }
	/*
	 * Now we get down to the real work.  The rules are as follows:
	 *
	 * (1) Starting with the year, see if a field in stime matches
	 *     a bit in the entry's time mask.
	 * (2) When a field is found that doesn't have a matched bit,
	 *     mark that field as the "field to increment".
	 * (3) As long as there is a field to increment, search in that field
	 *     for the next set bit.  If one is found, reset all less-
	 *     significant fields.  Otherwise, mark the next most-significant
	 *     field for incrementation.
	 */
	DBPRINT0(20, "within nextrun..\n");
	if (entp->ce_year[MASKBYTE (stimebuf.tm_year)]
	  & MASKBIT (stimebuf.tm_year))
	    {
	    if (entp->ce_month[MASKBYTE (stimebuf.tm_mon)]
	      & MASKBIT (stimebuf.tm_mon))
		{
		if ((entp->ce_moday[MASKBYTE (stimebuf.tm_mday)]
		    & MASKBIT (stimebuf.tm_mday))
		  &&  (entp->ce_wkday[MASKBYTE (stimebuf.tm_wday)]
		    & MASKBIT (stimebuf.tm_wday)))
		    {
		    if (entp->ce_hour[MASKBYTE (stimebuf.tm_hour)]
		      & MASKBIT (stimebuf.tm_hour))
			bumpfield = BUMPMINUTE; /* Must increment minute */
		    else
			bumpfield = BUMPHOUR; /* Must increment hour */
		    }
		else
		    bumpfield = BUMPDAY; /* Must increment day of month */
		}
	    else
		bumpfield = BUMPMONTH;	/* Must increment month */
	    }
	else
	    bumpfield = BUMPYEAR;	/* Must increment year */
	DBPRINT1(20, "bumpfield %d \n",bumpfield);
	while (bumpfield != BUMPNONE)
	    {
	    switch (bumpfield)
		{
		/*
		 * The year, month, and day must be calculated while keeping
		 * the weekday in mind.  This means that any attempt to bump
		 * them must be conditional, based on the possibility that the
		 * weekday might mismatch.
		 */
		case BUMPYEAR:
		    stimebuf.tm_year = nextset (stimebuf.tm_year,
		      entp->ce_year, YEARMAX);
		    stimebuf.tm_mon = -1;
		    if (stimebuf.tm_year < 0)
			{
			/*
			 * There are no more years.  If this is a cron
			 * job, return failure.  If it's an at job, we
			 * must have missed its execution time, so we
			 * will return the start time, causing the job
			 * to execute immediately.
			 */
			if (entp->ce_flags & EF_ATJOB)
			    return stime + 1;
			else
			    return 0;
			}
		    /* FALLTHRU */
		case BUMPMONTH:
		    stimebuf.tm_mon =
		      nextset (stimebuf.tm_mon, entp->ce_month, 12);
		    stimebuf.tm_mday = -1;
		    if (stimebuf.tm_mon < 0)
			{
			bumpfield = BUMPYEAR;
			break;
			}
		    /* FALLTHRU */
		case BUMPDAY:
		    stimebuf.tm_mday =
		      nextset (stimebuf.tm_mday, entp->ce_moday, 31);
		    stimebuf.tm_hour = -1;
		    if (stimebuf.tm_mday < 0
		      ||  stimebuf.tm_mday
			>= mlens[stimebuf.tm_mon]
			  + (leap (stimebuf.tm_year + yearbase)
			    &&  stimebuf.tm_mon == 1))
			{
			bumpfield = BUMPMONTH;
			break;
			}
		    stimebuf.tm_wday =
		      wkday (stimebuf.tm_year + yearbase - 1900,
			stimebuf.tm_mon, stimebuf.tm_mday + 1);
		    if ((entp->ce_wkday[MASKBYTE (stimebuf.tm_wday)]
		      & MASKBIT (stimebuf.tm_wday)) == 0)
			{		/* Wrong weekday, try again */
			stimebuf.tm_isdst = /* Just a handy scratch */
			  nextset (stimebuf.tm_wday, entp->ce_wkday, 7);
			if (stimebuf.tm_isdst > 0)
			    stimebuf.tm_mday +=
			      stimebuf.tm_isdst - stimebuf.tm_wday - 1;
			bumpfield = BUMPDAY;
			break;
			}
		    /* FALLTHRU */
		case BUMPHOUR:
		    stimebuf.tm_hour =
		      nextset (stimebuf.tm_hour, entp->ce_hour, 24);
		    stimebuf.tm_min = -1;
		    if (stimebuf.tm_hour < 0)
			{
			bumpfield = BUMPDAY;
			break;
			}
		    /* FALLTHRU */
		case BUMPMINUTE:
		    stimebuf.tm_min =
		      nextset (stimebuf.tm_min, entp->ce_min, 60);
		    if (stimebuf.tm_min < 0)
			bumpfield = BUMPHOUR;
		    else
			bumpfield = BUMPNONE; /* We're done! */
		    break;
		}
	    }
	/*
	 * Now convert stimebuf back to a time_t and return it.
	 */
	stimebuf.tm_year += yearbase - 1900;
	stimebuf.tm_mday++;
	stimebuf.tm_sec = 0;
	nexttime = tm_to_time (&stimebuf);
	if (nexttime != 0  &&  nexttime != (time_t) -1)
	  {
	    DBPRINT1(5, "next time is %ld before returning from next run\n",nexttime);
	    return nexttime;
	  }
	if (entp->ce_flags & EF_ATJOB)
	    {
	    /*
	     * This job is an "at" job, and has been scheduled for a
	     * nonexistent time (probably the "missing" hour during
	     * the daylight-savings switch).  Wildcard all of the
	     * fields at greater-than-minute resolution, which will
	     * cause this loop to find the next existing time with the
	     * same minutes field (e.g., 2:30 AM will become 3:30 AM,
	     * which is probably what the user will want).
	     */
	    (void) memset ((char *) entp->ce_month, ~0,
	      BITSIZE (12) * sizeof (MASKTYPE));
	    (void) memset ((char *) entp->ce_moday, ~0,
	      BITSIZE (31) * sizeof (MASKTYPE));
	    (void) memset ((char *) entp->ce_year, ~0,
	      BITSIZE (YEARMAX) * sizeof (MASKTYPE));
	    (void) memset ((char *) entp->ce_hour, ~0,
	      BITSIZE (24) * sizeof (MASKTYPE));
	    }
	}
    /* NOTREACHED */
    }

static int nextset (bit, mask, masksize) /* Find next set bit */
    reg2 int		bit;		/* Bit to start from */
    reg1 MASKTYPE *	mask;		/* Mask to search */
    reg3 int		masksize;	/* Number of bits in mask */
    {
    for (  ;  ++bit < masksize;  )
	if (mask[MASKBYTE (bit)] & MASKBIT (bit))
	    return bit;
    return -1;
    }
