#ifndef lint
static char Rcs_Id[] =
    "$Header$";
#endif

/*
 * $Log$
 * Revision 2.6  1998/07/30 20:42:30  geoff
 * Add longname support.
 *
 * Revision 2.5  1995/11/27  01:59:52  geoff
 * Fix a tiny spelling error in a comment.
 *
 * Revision 2.4  1995/05/11  05:05:59  geoff
 * Minor RCS header changes.
 *
 * Revision 2.3  86/11/29  19:54:41  geoff
 * Change all the register declarations to be portable across
 * machines with varying numbers of registers.
 * 
 * Revision 2.2  86/03/28  14:43:54  geoff
 * Fix some lint complaints.
 * 
 * Revision 2.1  86/01/04  15:31:45  geoff
 * Initial revision.
 * 
 */

#include "config.h"

char * tempfile ();			/* Create a temporoary file name */

static char base64[64] =		/* Base-64 conversion table */
    {
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
    'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
    'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
    'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
    'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
    'y', 'z', '-', ':'
    };

#ifdef LONGNAMES
#define MIN_X_COUNT	11		/* Minimum number of trailing X's */
#else /* LONGNAMES */
#define MIN_X_COUNT	6		/* Minimum number of trailing X's */
#endif /* LONGNAMES */

static unsigned long sequence = 0;	/* Sequence counter */

extern int	getpid ();		/* Return current process ID */

/*
 * Create a temporary file name.
 *
 * On a system with short (14-character) filenames, base points to a
 * modifiable area containing a prototype string that ends with at
 * least 6 X's.  The last 5 X's are filled in with the last 5 digits
 * of the decimal process ID, zero-filled.
 *
 * On a system with long filenames, base points to a modifiable area
 * containing a prototype string that ends with at least 11 X's.  The
 * last 10 X's are filled in with the last 10 digits of the decimal
 * process ID, zero-filled.
 *
 * Remaining X's are filled in with a base-64 encoding of a sequence
 * counter.  A check is made for duplications, and the filename is
 * changed if necessary.  However, there is a possible timing window
 * in the duplication check.
 */
char * tempfile (base)			/* Create a temporary file name */
    reg2 char *		base;		/* Where to store the result */
    {
    reg1 char *		fillp;		/* Pointer to where to fill in */
    reg4 char *		fill_e;		/* End (left hand side) of fill area */
    reg5 unsigned int	len;		/* Length of the string */
    reg6 unsigned long	pid;		/* Current process ID */
    reg3 unsigned long	seq;		/* For converting seq into base 64 */

    len = (unsigned) strlen (base);
    if (len < MIN_X_COUNT)
	return base;			/* Too short to do anything */
    /*
     * We will assume the last 6 or 11 characters are X's, even if
     * they aren't.  Tough luck if we're wrong.
     */
    pid = getpid ();
#ifdef LONGNAMES
    /*
     * Some compilers can't handle "long long" constants, so we can't
     * easily write portable code to limit the PID to 10 digits.
     * Rather than risk overrunning the input area, we'll do the
     * conversion by hand and limit it to 10 digits.  This assumes
     * MIN_X_COUNT is greater than 10.
     */
    fillp = base + len;
    fill_e = fillp - 10;
    while (fillp >= fill_e)
	{
	*--fillp = pid % 10 + '0';
	pid /= 10;
	}
#else /* LONGNAMES */
    fillp = base + len - 5;
    pid %= 100000;
    (void) sprintf (fillp, "%5.5d", pid);
     /*
     * Some sprintfs don't use the notation above for zero-filling, so
     * we'll replace blanks with zeros.
     */
    for (  ;  *fillp != '\0';  fillp++)
	{
	if (*fillp == ' ')
	    *fillp = '0';
	}
#endif /* LONGNAMES */
    /*
     * Produce the base-64 encoding of the sequence, and test for file
     * existence.  If none exists, break out of the loop.
     */
    fillp = base + len - (MIN_X_COUNT - 1);
    while (fillp >= base  &&  *--fillp == 'X')
	;
    fill_e = fillp + 1;
    do
	{
	for (fillp = base + len - (MIN_X_COUNT - 2), seq = sequence++;
	  fillp > fill_e;
	  )
	    {
	    *--fillp = base64[seq & 0x3F];
	    seq >>= 6;
	    }
	}
    while (access (base, 0) >= 0);	/* Loop while the file exists */
    return base;
    }
